"""
Pylint plugin: checks that feature toggles are properly annotated.
"""

from __future__ import absolute_import
import re

from pylint.checkers import BaseChecker, utils
from pylint.interfaces import IAstroidChecker

from .common import BASE_ID, check_visitors


def register_checkers(linter):
    """
    Register checkers.
    """
    linter.register_checker(FeatureToggleChecker(linter))

class AnnotationLines(object):
    """
    AnnotationLines provides utility methods to work with a string in terms of
    lines.  As an example, it can convert a Call node into a list of its contents
    separated by line breaks.
    """
    # Regex searches for annotations like: # .. toggle or # .. documented_elsewhere
    _ANNOTATION_REGEX = re.compile(r'[\s]*#[\s]*\.\.[\s]*(toggle|documented_elsewhere)')

    def __init__(self, module_node):
        """
        Arguments:
            module_node: The visited module node.
        """
        module_as_binary = module_node.stream().read()

        file_encoding = module_node.file_encoding
        if file_encoding is None:
            file_encoding = "UTF-8"

        module_as_string = module_as_binary.decode(file_encoding)
        self._list_of_string_lines = module_as_string.split("\n")

    def is_line_annotated(self, line_number):
        """
        Checks if the provided line number is annotated.
        """
        if line_number < 1 or self._line_count() < line_number:
            return False

        return bool(
            self._ANNOTATION_REGEX.match(self._get_line_contents(line_number))
        )

    def _line_count(self):
        """
        Gets the number of lines in the string.
        """
        return len(self._list_of_string_lines)

    def _get_line_contents(self, line_number):
        """
        Gets the line of text designated by the provided line number.
        """
        return self._list_of_string_lines[line_number - 1]


@check_visitors
class FeatureToggleChecker(BaseChecker):
    """
    Checks that feature toggles are properly annotated and best practices
    are followed.
    """
    __implements__ = (IAstroidChecker,)

    name = 'feature-toggle-checker'

    TOGGLE_NOT_ANNOTATED_MESSAGE_ID = 'feature-toggle-needs-doc'
    ILLEGAL_WAFFLE_MESSAGE_ID = 'illegal-waffle-usage'

    _CHECK_CAPITAL_REGEX = re.compile(r'[A-Z]')
    _WAFFLE_TOGGLE_CLASSES = ('WaffleFlag', 'WaffleSwitch', 'CourseWaffleFlag',)
    _ILLEGAL_WAFFLE_FUNCTIONS = ['flag_is_active', 'switch_is_active',]

    msgs = {
        'E%d40' % BASE_ID: (
            u"feature toggle (%s) is missing annotation",
            TOGGLE_NOT_ANNOTATED_MESSAGE_ID,
            "feature toggle is missing annotation",
        ),
        'E%d41' % BASE_ID: (
            u"illegal waffle usage with (%s): use utility classes {}.".format(', '.join(_WAFFLE_TOGGLE_CLASSES)),
            ILLEGAL_WAFFLE_MESSAGE_ID,
            u"illegal waffle usage: use utility classes {}.".format(', '.join(_WAFFLE_TOGGLE_CLASSES)),
        ),
    }

    def __init__(self, *args, **kwargs):
        super(FeatureToggleChecker, self).__init__(*args, **kwargs)
        self._lines = None

    def visit_module(self, node):
        """Parses the module code to provide access to comments."""
        self._lines = AnnotationLines(node)

    def check_waffle_class_annotated(self, node):
        """
        Check Call node for waffle class instantiation with missing annotations.
        """
        if not hasattr(node.func, 'name'):
            return

        # Looking for class instantiation, so should start with a capital letter
        starts_with_capital = self._CHECK_CAPITAL_REGEX.match(node.func.name)
        if not starts_with_capital:
            return

        # Search for toggle classes that require an annotation
        if not node.func.name.endswith(self._WAFFLE_TOGGLE_CLASSES):
            return

        if not self._lines.is_line_annotated(node.lineno - 1):
            feature_toggle_name = 'UNKNOWN'

            if node.keywords is not None:
                for node_key in node.keywords:
                    if node_key.arg == "flag_name":
                        feature_toggle_name = node_key.value.value

            if feature_toggle_name == 'UNKNOWN':
                if len(node.args) >= 2:
                    feature_toggle_name = node.args[1].as_string()

            self.add_message(
                self.TOGGLE_NOT_ANNOTATED_MESSAGE_ID,
                args=(feature_toggle_name,),
                node=node,
            )

    def check_configuration_model_annotated(self, node):
        """
        Checks class definitions to see if they subclass ConfigurationModel.
        If they do, they should be correctly annotated.
        """
        if "ConfigurationModel" not in node.basenames:
            return
        if not self._lines.is_line_annotated(node.lineno - 1):
            config_model_subclass_name = node.name

            self.add_message(
                self.TOGGLE_NOT_ANNOTATED_MESSAGE_ID,
                args=(config_model_subclass_name,),
                node=node,
            )

    def check_django_feature_flag_annotated(self, node):
        """
        Checks dictionary definitions to see if the django feature flags
        dict FEATURES is being set. If it is, entries should be
        correctly annotated.
        """
        try:
            parent_target_name = node.parent.targets[0].name
        except AttributeError:
            return

        if parent_target_name == "FEATURES":
            for key, _ in node.items:
                if not self._lines.is_line_annotated(key.lineno - 1):
                    django_feature_toggle_name = key.value

                    self.add_message(
                        self.TOGGLE_NOT_ANNOTATED_MESSAGE_ID,
                        args=(django_feature_toggle_name,),
                        node=node,
                    )

    def check_illegal_waffle_usage(self, node):
        """
        Check Call node for illegal waffle calls.
        """
        if not hasattr(node.func, 'name'):
            return

        if node.func.name in self._ILLEGAL_WAFFLE_FUNCTIONS:
            feature_toggle_name = 'UNKNOWN'
            if len(node.args) >= 1:
                feature_toggle_name = node.args[0].as_string()

            self.add_message(
                self.ILLEGAL_WAFFLE_MESSAGE_ID,
                args=(feature_toggle_name,),
                node=node,
            )

    @utils.check_messages(TOGGLE_NOT_ANNOTATED_MESSAGE_ID, ILLEGAL_WAFFLE_MESSAGE_ID)
    def visit_call(self, node):
        """
        Performs various checks on Call nodes.
        """
        self.check_waffle_class_annotated(node)
        self.check_illegal_waffle_usage(node)

    @utils.check_messages(TOGGLE_NOT_ANNOTATED_MESSAGE_ID)
    def visit_classdef(self, node):
        """
        Checks class definitions for potential ConfigurationModel
        implementations.
        """
        self.check_configuration_model_annotated(node)

    @utils.check_messages(TOGGLE_NOT_ANNOTATED_MESSAGE_ID)
    def visit_dict(self, node):
        """
        Checks Dict nodes in case a Django FEATURES dictionary is being
        initialized.
        """
        self.check_django_feature_flag_annotated(node)



"""edx_lint pylint module.

Add this to your pylintrc::

    load-plugins=edx_lint.pylint

"""

from __future__ import absolute_import
from edx_lint.pylint import plugin

# Pylint both insists on, and complains about, this name :)
# pylint: disable=invalid-name
register = plugin.register



"""Check that strings are unicode enough."""

import ast
import json
import re
import token

import six

from pylint.checkers import BaseTokenChecker
from pylint.interfaces import ITokenChecker, IRawChecker

from .common import BASE_ID, check_visitors


def register_checkers(linter):
    """Register checkers."""
    linter.register_checker(UnicodeFormatStringChecker(linter))


@check_visitors
class UnicodeFormatStringChecker(BaseTokenChecker):
    """
    Checks that strings are unicode.

    Message ID is: unicode-format-string
    """
    # ITokenChecker gets us process_tokens support.
    # IRawChecker gets us process_module support.
    __implements__ = (ITokenChecker, IRawChecker)

    name = 'unicode-format-string-checker'

    MESSAGE_ID = 'unicode-format-string'
    msgs = {
        'C%d11' % BASE_ID: (
            "Human-readable format strings should be unicode",
            MESSAGE_ID,
            "Human-readable format strings should be unicode",
        ),
    }

    def __init__(self, *args, **kwargs):
        super(UnicodeFormatStringChecker, self).__init__(*args, **kwargs)
        self._unicode_literals = False

    def process_module(self, node):
        """Called for each module being examined."""
        self._unicode_literals = "unicode_literals" in node.future_imports

    def process_tokens(self, tokens):
        if six.PY3:
            # Python 3 doesn't have this issue.
            return

        if self._unicode_literals:
            # If the whole module is __future__'d, then it's all fine.
            return

        previous_line_has_definition = False
        current_line_has_definition = False
        first_non_comment_line = None
        for index, (tok_type, tok_text, start, _, _) in enumerate(tokens):
            if first_non_comment_line is None and tok_type != 53: # 53 is token.COMMENT
                first_non_comment_line = index

            # Keep track of whether previous line contained a function or class definition
            if tok_type == token.NEWLINE or tok_type == 54:  # 54 is token.NL
                previous_line_has_definition = current_line_has_definition
                current_line_has_definition = False
            elif tok_text in ['def', 'class']:
                current_line_has_definition = True

            if tok_type == token.STRING:
                if tok_text.lower().startswith(("u", "b", "ur", "br")):
                    # An explicit prefix is fine.
                    continue

                value = ast.literal_eval(tok_text)  # TODO: can this fail?

                if " " not in value:
                    # If there's no space, then it's probably not a message,
                    # leave it alone.
                    continue

                if not re.search(r"[%{]", value):
                    # There's no formatting character, leave it alone.
                    continue

                if previous_line_has_definition:
                    # Previous line probably contains a function or class definition
                    # so this string is probably a docstring
                    continue
                if first_non_comment_line and index == first_non_comment_line:
                    # If this string is the first line of the file (excluding comments) it is a docstring
                    continue

                try:
                    json.loads(value)
                except ValueError:
                    # It's invalid JSON, keep checking.
                    pass
                else:
                    # Valid JSON isn't human-readable, leave it alone.
                    continue    # pragma: no cover   grr.. https://github.com/nedbat/coveragepy/issues/198

                self.add_message(self.MESSAGE_ID, line=start[0])



"""The edx_lint check command."""
from __future__ import print_function

import os

from edx_lint.tamper_evident import TamperEvidentFile


def check_main(argv):
    """
    check FILENAME
        Check that FILENAME has not been edited since writing.
    """
    if len(argv) != 1:
        print("Please provide the name of a file to check.")
        return 1

    filename = argv[0]

    if os.path.exists(filename):
        print(u"Checking existing copy of %s" % filename)
        tef = TamperEvidentFile(filename)
        if tef.validate():
            print(u"Your copy of %s is good" % filename)
        else:
            print(u"Your copy of %s seems to have been edited" % filename)
    else:
        print(u"You don't have a copy of %s" % filename)

    return 0



"""The edx_lint write command."""

from __future__ import print_function

import os

import six
from six.moves import configparser
from six.moves import cStringIO

import pkg_resources

from edx_lint.tamper_evident import TamperEvidentFile
from edx_lint.configfile import merge_configs


WARNING_HEADER = u"""\
# ***************************
# ** DO NOT EDIT THIS FILE **
# ***************************
#
# This file was generated by edx-lint: http://github.com/edx/edx-lint
#
# If you want to change this file, you have two choices, depending on whether
# you want to make a local change that applies only to this repo, or whether
# you want to make a central change that applies to all repos using edx-lint.
#
# LOCAL CHANGE:
#
# 1. Edit the local {tweaks_name} file to add changes just to this
#    repo's file.
#
# 2. Run:
#
#       $ edx_lint write {filename}
#
# 3. This will modify the local file.  Submit a pull request to get it
#    checked in so that others will benefit.
#
#
# CENTRAL CHANGE:
#
# 1. Edit the {filename} file in the edx-lint repo at
#    https://github.com/edx/edx-lint/blob/master/edx_lint/files/{filename}
#
# 2. install the updated version of edx-lint (in edx-lint):
#
#       $ pip install .
#
# 3. Run (in edx-lint):
#
#       # uses {tweaks_name} from edx-lint for linting in edx-lint
#       # NOTE: Use Python 3.x, which no longer includes comments in the output file
#       $ edx_lint write {filename}
#
# 4. Make a new version of edx_lint, submit and review a pull request with the
#    {filename} update, and after merging, update the edx-lint version by
#    creating a new tag in the repo (uses pbr).
#
# 5. In your local repo, install the newer version of edx-lint.
#
# 6. Run:
#
#       # uses local {tweaks_name}
#       $ edx_lint write {filename}
#
# 7. This will modify the local file.  Submit a pull request to get it
#    checked in so that others will benefit.
#
#
#
#
#
# STAY AWAY FROM THIS FILE!
#
#
#
#
#
# SERIOUSLY.
#
# ------------------------------
"""

def write_main(argv):
    """
    write FILENAME
        Write a local copy of FILENAME using FILENAME_tweaks for local tweaks.
    """
    if len(argv) != 1:
        print("Please provide the name of a file to write.")
        return 1

    filename = argv[0]
    resource_name = "files/" + filename
    tweaks_name = amend_filename(filename, "_tweaks")

    if not pkg_resources.resource_exists("edx_lint", resource_name):
        print(u"Don't have file %r to write." % filename)
        return 2

    if os.path.exists(filename):
        print(u"Checking existing copy of %s" % filename)
        tef = TamperEvidentFile(filename)
        if not tef.validate():
            bak_name = amend_filename(filename, "_backup")
            print(u"Your copy of %s seems to have been edited, renaming it to %s" % (filename, bak_name))
            if os.path.exists(bak_name):
                print(u"A previous %s exists, deleting it" % bak_name)
                os.remove(bak_name)
            os.rename(filename, bak_name)

    print(u"Reading edx_lint/files/%s" % filename)
    cfg = configparser.RawConfigParser()
    resource_string = pkg_resources.resource_string("edx_lint", resource_name).decode("utf8")

    # pkg_resources always reads binary data (in both python2 and python3).
    # ConfigParser.read_string only exists in python3, so we have to wrap the string
    # from pkg_resources in a cStringIO so that we can pass it into ConfigParser.readfp.
    if six.PY2:
        cfg.readfp(cStringIO(resource_string), resource_name)
    else:
        cfg.read_string(resource_string, resource_name)     # pylint: disable=no-member

    if os.path.exists(tweaks_name):
        print(u"Applying local tweaks from %s" % tweaks_name)
        cfg_tweaks = configparser.RawConfigParser()
        cfg_tweaks.read([tweaks_name])

        merge_configs(cfg, cfg_tweaks)

    print(u"Writing %s" % filename)
    output_text = cStringIO()
    output_text.write(WARNING_HEADER.format(filename=filename, tweaks_name=tweaks_name))
    cfg.write(output_text)

    out_tef = TamperEvidentFile(filename)
    if six.PY2:
        output_bytes = output_text.getvalue()
    else:
        output_bytes = output_text.getvalue().encode("utf8")
    out_tef.write(output_bytes)

    return 0


def amend_filename(filename, amend):
    """Amend a filename with a suffix.

    amend_filename("foo.txt", "_tweak") --> "foo_tweak.txt"

    """
    base, ext = os.path.splitext(filename)
    amended_name = base + amend + ext
    return amended_name



"""
Test ignore-long-lines regex
"""

from .pylint_test import run_pylint


def test_ignore_long_lines():
    source = """\
        # Bad
        a = 1  # hello world hello world hello world hello world hello world hello world hello world hello world hello world hello world
        # ..pii: hello world hello world hello world hello world hello world hello world hello world hello world hello world hello world
        '''
        < https://www.test.com/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world>
        hello world hello world hello world hello world hello world hello world hello world hello world hello world hello world hello world
        '''

        # Good
        '''
        .. pii: hello world hello world hello world hello world hello world hello world hello world hello world hello world hello world
        '''
        # .. pii: hello world hello world hello world hello world hello world hello world hello world hello world hello world hello world
        '''
        <https://www.test.com/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world>
        https://www.test.com/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world
        '''
        # <https://www.test.com/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world>
        # https://www.test.com/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world/hello/world
        # .. toggle_description: this is a good toggle, it goes back and forth, toggling a lot, which is what we expect toggles to do.
    """
    msg_ids = "line-too-long"
    messages = run_pylint(source, msg_ids)

    # normalize the messages by stripping off the 3rd segment.
    messages = {
        ':'.join(message.split(':')[:2])
        for message in messages
    }

    expected = {
        '2:line-too-long',
        '3:line-too-long',
        '5:line-too-long',
        '6:line-too-long',
    }
    assert expected == messages



"""
Tests of the lint-amnesty command.
"""

from collections import defaultdict
from io import StringIO
import itertools
import textwrap
import unittest

from edx_lint.cmd.amnesty import PylintError, fix_pylint, PYLINT_EXCEPTION_REGEX, parse_pylint_output

class AmnestyTest(unittest.TestCase):
    """
    Tests of lint-amnesty.
    """

    def assert_amnesty(self, input_code, errors, expected):
        """
        Assert that fix_pylint produces ``expected`` when fed ``input_code`` and the
        list of errors ``errors``.

        Arguments:
            input_code: A string of python code. Will be textwrap.dedented.
            errors: A list of PylintErrors
            expected: A string of python code. Will be textwrap.dedented.
        """
        input_code = textwrap.dedent(input_code)
        expected = textwrap.dedent(expected)

        errors_by_line = defaultdict(list)
        for error in errors:
            errors_by_line[error.linenum].append(error)

        output_lines = itertools.chain.from_iterable(
            fix_pylint(line, errors_by_line[lineno])
            for lineno, line
            in enumerate(StringIO(input_code), start=1)
        )

        self.assertEqual(expected.split(u'\n'), "".join(output_lines).split(u'\n'))

    def assert_pylint_exception_match(self, expected, line):
        """
        Assert that PYLINT_EXCEPTION_REGEX mateches ``expected`` in ``line``.
        """
        self.assertEqual(
            expected,
            PYLINT_EXCEPTION_REGEX.search(line).group(0)
        )

    def test_pylint_exception_re(self):
        self.assert_pylint_exception_match(
            "  # pylint: disable=some-exception, other-exception",
            "some line  # pylint: disable=some-exception, other-exception",
        )
        self.assert_pylint_exception_match(
            "  # pylint: disable=some-exception,other-exception",
            "some line  # pylint: disable=some-exception,other-exception",
        )
        self.assert_pylint_exception_match(
            "  # pylint: disable=some-exception",
            "some line  # pylint: disable=some-exception",
        )
        self.assert_pylint_exception_match(
            "  # pylint: disable=some-exception",
            "some line  # pylint: disable=some-exception#noqa",
        )
        self.assert_pylint_exception_match(
            "  # pylint: disable=some-exception",
            "some line  # pylint: disable=some-exception  #noqa",
        )
        self.assert_pylint_exception_match(
            "# pylint: disable=some-exception",
            "some line# pylint: disable=some-exception#noqa",
        )

    def test_parse_pylint_output(self):
        pylint_output = textwrap.dedent('''\
            ************* Module cms.urls
            cms/urls.py:1: [C0111(missing-docstring), ] Missing module docstring
            ************* Module cms.envs.test
            cms/envs/test.py:81: [C0301(line-too-long), ] Line too long (144/120)
            ************* Module test.test_amnesty
            test/test_amnesty.py:78: [C0326(bad-whitespace), ] Exactly one space required around assignment
                    pylint_output=textwrap.dedent("""\\
                                 ^
        ''')
        self.assertEqual(
            [
                PylintError('cms/urls.py', 1, 'C0111', 'missing-docstring', '', 'Missing module docstring'),
                PylintError('cms/envs/test.py', 81, 'C0301', 'line-too-long', '', 'Line too long (144/120)'),
                PylintError(
                    'test/test_amnesty.py', 78, 'C0326', 'bad-whitespace', '',
                    "Exactly one space required around assignment"
                ),
            ],
            list(parse_pylint_output(pylint_output.split('\n'))),
        )

    def test_add_supression(self):
        input_code = u"""\
            def func(arg):
                pass
        """

        errors = [
            PylintError('test.py', 1, 'W0613', 'unused-argument', 'func', "Unused argument 'arg'")
        ]

        expected = u"""\
            def func(arg):  # lint-amnesty, pylint: disable=unused-argument
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_add_multiple_suppressions_same_line(self):
        input_code = u"""\
            def func(arg, arg):
                pass
        """

        errors = [
            PylintError('test.py', 1, 'W0613', 'unused-argument', 'func', "Unused argument 'arg'"),
            PylintError('test.py', 1, 'E0108', 'duplicate-argument-name', 'func', "Duplicate argument name 'arg'")
        ]

        expected = u"""\
            def func(arg, arg):  # lint-amnesty, pylint: disable=duplicate-argument-name, unused-argument
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_add_to_existing_suppression(self):
        input_code = u"""\
            def func(arg, arg):  # pylint: disable=unused-argument
                pass
        """

        errors = [
            PylintError('test.py', 1, 'E0108', 'duplicate-argument-name', 'func', "Duplicate argument name 'arg'")
        ]

        expected = u"""\
            def func(arg, arg):  # lint-amnesty, pylint: disable=duplicate-argument-name, unused-argument
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_add_to_existing_suppression_trailing_comment(self):
        input_code = u"""\
            def func(arg, arg):  # pylint: disable=unused-argument  # noqa
                pass
        """

        errors = [
            PylintError('test.py', 1, 'E0108', 'duplicate-argument-name', 'func', "Duplicate argument name 'arg'")
        ]

        expected = u"""\
            def func(arg, arg):  # lint-amnesty, pylint: disable=duplicate-argument-name, unused-argument  # noqa
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_add_suppression_trailing_comment(self):
        input_code = u"""\
            def func(arg, arg):  # noqa
                pass
        """

        errors = [
            PylintError('test.py', 1, 'E0108', 'duplicate-argument-name', 'func', "Duplicate argument name 'arg'")
        ]

        expected = u"""\
            def func(arg, arg):  # noqa  # lint-amnesty, pylint: disable=duplicate-argument-name
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_remove_useless_suppression(self):
        input_code = u"""\
            def func(arg, arg):  # pylint: disable=unused-argument
                pass
        """

        errors = [
            PylintError(
                'test.py', 1, 'I0021', 'useless-suppression', 'func',
                "Useless suppression of 'unused-argument'"
            )
        ]

        expected = u"""\
            def func(arg, arg):
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_remove_useless_suppression_leave_rest(self):
        input_code = u"""\
            def func(arg, arg):  # pylint: disable=unused-argument, duplicate-argument-name
                pass
        """

        errors = [
            PylintError(
                'test.py', 1, 'I0021', 'useless-suppression', 'func',
                "Useless suppression of 'unused-argument'"
            )
        ]

        expected = u"""\
            def func(arg, arg):  # pylint: disable=duplicate-argument-name
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_replace_useless_suppression(self):
        input_code = u"""\
            def func(arg, arg):  # pylint: disable=unused-argument
                pass
        """

        errors = [
            PylintError(
                'test.py', 1, 'I0021', 'useless-suppression', 'func',
                "Useless suppression of 'unused-argument'"
            ),
            PylintError(
                'test.py', 1, 'E0108', 'duplicate-argument-name', 'func',
                "Duplicate argument name 'arg'"
            ),
        ]

        expected = u"""\
            def func(arg, arg):  # lint-amnesty, pylint: disable=duplicate-argument-name
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_fix_multiple_lines(self):
        input_code = u"""\
            def func_a(arg):
                pass

            def func_b(arg, arg):
                pass
        """

        errors = [
            PylintError('test.py', 1, 'W0613', 'unused-argument', 'func', "Unused argument 'arg'"),
            PylintError('test.py', 4, 'E0108', 'duplicate-argument-name', 'func', "Duplicate argument name 'arg'"),
        ]

        expected = u"""\
            def func_a(arg):  # lint-amnesty, pylint: disable=unused-argument
                pass

            def func_b(arg, arg):  # lint-amnesty, pylint: disable=duplicate-argument-name
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_fix_module_docstring_missing(self):
        input_code = u"""\
            def func(arg):
                pass
        """

        errors = [
            PylintError('test.py', 1, 'C0111', 'missing-docstring', 'func', "Missing module docstring"),
        ]

        expected = u"""\
            # lint-amnesty, pylint: disable=missing-docstring
            def func(arg):
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_fix_module_and_func_docstring_missing(self):
        input_code = u"""\
            def func(arg):
                pass
        """

        errors = [
            PylintError('test.py', 1, 'C0111', 'missing-docstring', 'func', "Missing module docstring"),
            PylintError('test.py', 1, 'C0111', 'missing-docstring', 'func', "Missing function docstring"),
        ]

        expected = u"""\
            # lint-amnesty, pylint: disable=missing-docstring
            def func(arg):  # lint-amnesty, pylint: disable=missing-docstring
                pass
        """

        self.assert_amnesty(input_code, errors, expected)

    def test_fix_many_errors(self):
        input_code = u"""\
            def foo(self, arg, blarg, flarg, blanth, notehu, onethu, psonatehu):
                pass"""

        errors = [
            PylintError('lint_test.py', 2, 'C0304', 'missing-final-newline', '', 'Final newline missing'),
            PylintError('lint_test.py', 1, 'C0111', 'missing-docstring', '', 'Missing module docstring'),
            PylintError('lint_test.py', 1, 'C0102', 'blacklisted-name', 'foo', 'Black listed name "foo"'),
            PylintError('lint_test.py', 1, 'C0111', 'missing-docstring', 'foo', 'Missing function docstring'),
            PylintError('lint_test.py', 1, 'W0613', 'unused-argument', 'foo', "Unused argument 'notehu'"),
            PylintError('lint_test.py', 1, 'W0613', 'unused-argument', 'foo', "Unused argument 'self'"),
            PylintError('lint_test.py', 1, 'W0613', 'unused-argument', 'foo', "Unused argument 'psonatehu'"),
            PylintError('lint_test.py', 1, 'W0613', 'unused-argument', 'foo', "Unused argument 'blanth'"),
            PylintError('lint_test.py', 1, 'W0613', 'unused-argument', 'foo', "Unused argument 'onethu'"),
            PylintError('lint_test.py', 1, 'W0613', 'unused-argument', 'foo', "Unused argument 'arg'"),
            PylintError('lint_test.py', 1, 'W0613', 'unused-argument', 'foo', "Unused argument 'blarg'"),
            PylintError('lint_test.py', 1, 'W0613', 'unused-argument', 'foo', "Unused argument 'flarg'"),
        ]

        expected = u"""\
            # lint-amnesty, pylint: disable=missing-docstring
            def foo(self, arg, blarg, flarg, blanth, notehu, onethu, psonatehu):  # lint-amnesty, pylint: disable=blacklisted-name, missing-docstring, unused-argument
                pass  # lint-amnesty, pylint: disable=missing-final-newline"""

        self.assert_amnesty(input_code, errors, expected)



"""Test unicode_check.py"""

import six

from .pylint_test import run_pylint


MSG_IDS = "unicode-format-string"


def test_unicode_checker():
    source = """\
    class Test():
        "docstring this is {fine}"

        def test():
            "docstring this is {fine}"
            pass
    "this is fine"
    "this is not {}".format("fine")         #=A
    "This is not %s" % ("fine",)            #=B
    '{"json": "is fine"}'
    u"Unicode strings are {}".format("fine")
    b"byte strings are %s." % ("fine,")
    "onewordis{}".format("fine")
    "pragma makes this {}".format("fine")   # pylint: disable=unicode-format-string
    """
    messages = run_pylint(source, MSG_IDS)

    # This checker only makes messages on Python 2.
    if six.PY2:
        expected = {
            'A:unicode-format-string:Human-readable format strings should be unicode',
            'B:unicode-format-string:Human-readable format strings should be unicode',
            }
    else:
        expected = set()

    assert expected == messages


def test_unicode_checker_with_future():
    source = """\
        "This is a docstring"
        from __future__ import unicode_literals

        "this is fine"
        u"this is fine"
        "this is {}".format("fine")
        "This is %s" % ("fine",)
        '{"json": "is fine"}'
        u"Unicode strings are {}".format("fine")
        "onewordis{}".format("fine")
        """
    messages = run_pylint(source, MSG_IDS)
    assert not messages



""" Tests for the command line executable. """

import os
import shutil
import tempfile
import unittest

from edx_lint.cmd import main


PYLINTRC = 'pylintrc'
PYLINTRC_TWEAKS = 'pylintrc_tweaks'
PYLINTRC_BACKUP = 'pylintrc_backup'

class WriteCommandTest(unittest.TestCase):
    """ Tests for the write command. """

    def setUp(self):
        super(WriteCommandTest, self).setUp()

        # Make a temporary directory to work in, and rm-rf it when we're done.
        tempdir = tempfile.mkdtemp(suffix="_edx_lint_test")
        self.addCleanup(shutil.rmtree, tempdir)

        # Change into the temp directory, and change back when we are done.
        thisdir = os.getcwd()
        self.addCleanup(os.chdir, thisdir)
        os.chdir(tempdir)

        # Assure the file does not already exist.
        self.assert_not_file(PYLINTRC)

    def call_command(self, argv=None):
        """ Call an edx_lint script command.

        Arguments:
            argv (list) -- arguments to pass to the edx_lint script
        """
        return main.main(argv)

    def assert_file(self, filename, contains=None, not_contains=None):
        """Assert that a file exists, and optionally, contains some text."""
        self.assertTrue(os.path.isfile(filename))
        if contains is not None or not_contains is not None:
            with open(filename) as f:
                text = f.read()
                if contains is not None:                    # pragma: no branch
                    self.assertIn(contains, text)
                if not_contains is not None:                # pragma: no branch
                    self.assertNotIn(not_contains, text)

    def assert_not_file(self, filename):
        """Assert that a file doesn't exist."""
        self.assertFalse(os.path.isfile(filename))

    def test_write_arg_errors(self):
        self.assertEqual(1, self.call_command(['write']))
        self.assertEqual(2, self.call_command(['write', 'xyzzy']))

    def test_write_creates_pylintrc(self):
        """ Verify the command writes a pylintrc file. """
        # Create the pylintrc.
        self.assertEqual(0, self.call_command(['write', PYLINTRC]))
        self.assert_file(PYLINTRC)

        # Try to create it again, it will verify it.
        self.assertEqual(0, self.call_command(['write', PYLINTRC]))

    def test_write_backups_modified_file(self):
        # Make the pylintrc.
        self.assertEqual(0, self.call_command(['write', PYLINTRC]))
        self.assert_file(PYLINTRC)
        self.assert_not_file(PYLINTRC_BACKUP)

        # Modify it.
        with open(PYLINTRC, "a") as pylintrc:
            pylintrc.write("# modified!\n")

        # Try to write again.
        self.assertEqual(0, self.call_command(['write', PYLINTRC]))
        self.assert_file(PYLINTRC_BACKUP, contains="# modified!")

        # And if we write again, the previous backup is deleted.
        with open(PYLINTRC, "a") as pylintrc:
            pylintrc.write("# changed!\n")
        self.assertEqual(0, self.call_command(['write', PYLINTRC]))
        self.assert_file(PYLINTRC_BACKUP, contains="# changed!", not_contains="# modified!")

    def test_write_applies_tweaks(self):
        # Create a tweaks file.
        with open(PYLINTRC_TWEAKS, "w") as tweaks:
            tweaks.write("[MASTER]\nxyzzy = plugh\n")

        # Write the pylintrc.
        self.assertEqual(0, self.call_command(['write', PYLINTRC]))

        # The tweak is in there!
        self.assert_file(PYLINTRC, contains="xyzzy = plugh")



"""Test super_check.py"""

import pytest

from .pylint_test import run_pylint

MSG_IDS = "super-method-not-called,non-parent-method-called"

@pytest.mark.parametrize("method", [
    "setUp",
    "tearDown",
    "setUpClass",
    "tearDownClass",
])
def test_unittest_super_check(method):
    # pylint: disable=unicode-format-string
    source = """\
        import unittest

        class GoodTestCase(unittest.TestCase):
            def {method}(self):
                super(GoodTestCase, self).{method}()

        class BadTestCase(unittest.TestCase):
            def {method}(self):         #=A
                self.i_am_bad = True

        class OldSchool(Mixin, unittest.TestCase):
            def {method}(self):
                Mixin.{method}(self)
                unittest.TestCase.{method}(self)

        def {method}(xyzzy):
            # Weird, but who cares?
            pass
        """.format(method=method)
    messages = run_pylint(source, MSG_IDS)
    expected = {"A:super-method-not-called:super(...).{}() not called (unittest.case.TestCase)".format(method)}
    assert expected == messages


@pytest.mark.parametrize("method", [
    "setUpTestData",
])
def test_django_super_check(method):
    # pylint: disable=unicode-format-string
    source = """\
        import django

        class GoodTestCase(django.test.TestCase):
            def {method}(self):
                super(GoodTestCase, self).{method}()

        class BadTestCase(django.test.TestCase):
            def {method}(self):         #=A
                self.i_am_bad = True
        """.format(method=method)
    messages = run_pylint(source, MSG_IDS)
    expected = {"A:super-method-not-called:super(...).setUpTestData() not called (django.test.testcases.TestCase)"}
    assert expected == messages


def test_hamfisted_super():
    source = """\
        import unittest

        class SomeOtherClass(object):
            def setUp(self):
                pass

        class BadTestCase(unittest.TestCase):
            def setUp(self):
                foo("What").setUp(self)
                What.setUp(self)
                SomeOtherClass.setUp(self)   #=A
                super(BadTestCase, self).setUp()
        """
    messages = run_pylint(source, MSG_IDS)
    expected = {"A:non-parent-method-called:setUp() was called from a non-parent class (source.SomeOtherClass)"}
    assert expected == messages


def test_good_super():
    source = """\
        import unittest

        def foo_func(bar):
            pass

        class GoodTestCase(unittest.TestCase):
            def setUp(self):
                self.foo_meth(bar)
                foo_func(bar)
                super(GoodTestCase, self).setUp()

            def tearDown(self):
                # Not sure why you would do it this way, ...
                base = super(GoodTestCase, self)
                base.tearDown()

            def foo_meth(self, bar):
                foo_func(bar)

            def test_something(self):
                pass

        class NotSureWhatThisIs(object):
            def setUp(self):
                pass
        """
    messages = run_pylint(source, MSG_IDS)
    assert not messages



"""
edx-lint
========

A collection of code quality tools:

- A few pylint plugins to check for quality issues pylint misses.

- A command-line tool to generate config files like pylintrc from a master
  file (part of edx_lint), and a repo-specific tweaks file.

"""

import os

from setuptools import setup

# pbr does some things we don't need. Turn them off the only way pbr gives us.
os.environ['SKIP_GENERATE_AUTHORS'] = '1'
os.environ['SKIP_WRITE_GIT_CHANGELOG'] = '1'

setup(
    # to update the version (uses pbr), create a new version tag in the repo
    setup_requires=['pbr>=1.9', 'setuptools>=17.1'],
    pbr=True,
)



"""
Checks for unsafe ``yaml.load()`` calls.
The PyYAML library has not yet released a version
that addresses the vulnerability (listed below) in the commonly
called ``load()`` function.

Vulnerability: https://nvd.nist.gov/vuln/detail/CVE-2017-18342
PyYAML release plan: https://github.com/yaml/pyyaml/issues/193

When a new release is published, we may be able to remove this check.
"""

from pylint.checkers import BaseChecker, utils
from pylint.interfaces import IAstroidChecker

from .common import BASE_ID, check_visitors


def register_checkers(linter):
    """Register checkers."""
    linter.register_checker(YamlLoadChecker(linter))


@check_visitors
class YamlLoadChecker(BaseChecker):
    """
    Checks for unsafe ``yaml.load()`` calls.
    """

    __implements__ = (IAstroidChecker,)

    name = 'yaml-load-checker'

    MESSAGE_ID = 'unsafe-yaml-load'

    UNSAFE_CALLS = {'yaml.load', 'yaml.load_all'}

    msgs = {
        'C{}57'.format(BASE_ID): (
            u'yaml.load%s() call is unsafe, use yaml.safe_load%s()',
            MESSAGE_ID,
            'yaml.load*() is unsafe',
        ),
    }

    @utils.check_messages(MESSAGE_ID)
    def visit_call(self, node):
        func_name = node.func.as_string()
        if func_name in self.UNSAFE_CALLS:
            suffix = func_name.lstrip('yaml.load')
            self.add_message(self.MESSAGE_ID, args=(suffix, suffix), node=node)
